#include "mesh.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdio>

using namespace std;
using namespace Eigen;
using namespace surface_mesh;

Mesh::~Mesh()
{
    if(_ready){
        glDeleteBuffers(1, &_indicesBuffer);
        glDeleteBuffers(2, _vbo);
        glDeleteVertexArrays(1,&_vao);
    }
}

void Mesh::load(const string& filename)
{
    cerr << "Loading: " << filename << endl;

    _halfEdge.read(filename);
    load();
}

void Mesh::load()
{
    if(!_halfEdge.is_triangle_mesh())
        _halfEdge.triangulate();

    _halfEdge.update_face_normals();
    _halfEdge.update_vertex_normals();

    _positions.resize(0);
    _normals.resize(0);
    _positions.reserve(_halfEdge.vertices_size());
    _normals.reserve(_halfEdge.vertices_size());

    // vertex properties
    Surface_mesh::Vertex_property<Point> vertices = _halfEdge.get_vertex_property<Point>("v:point");
    Surface_mesh::Vertex_property<Point> vnormals = _halfEdge.get_vertex_property<Point>("v:normal");

    // vertex iterator
    Surface_mesh::Vertex_iterator vit;

    for(vit = _halfEdge.vertices_begin(); vit != _halfEdge.vertices_end(); ++vit)
    {
        _positions.push_back(vertices[*vit]);
        _normals.push_back(vnormals[*vit]);
        _bbox.extend(vertices[*vit]);
    }

    _indices.resize(0);

    // face iterator
    Surface_mesh::Face_iterator fit, fend = _halfEdge.faces_end();
    // vertex circulator
    Surface_mesh::Vertex_around_face_circulator fvit, fvend;
    Surface_mesh::Vertex v0, v1, v2;
    for (fit = _halfEdge.faces_begin(); fit != fend; ++fit)
    {
        fvit = fvend = _halfEdge.vertices(*fit);
        v0 = *fvit;
        ++fvit;
        v2 = *fvit;

        do{
            v1 = v2;
            ++fvit;
            v2 = *fvit;
            _indices.push_back(Vector3i(v0.idx(), v1.idx(), v2.idx()));
        } while (++fvit != fvend);
    }
}

void Mesh::init()
{
    glGenVertexArrays(1, &_vao);
    glGenBuffers(2, _vbo);

    glBindVertexArray(_vao);

    glBindBuffer(GL_ARRAY_BUFFER, _vbo[0]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vector3f)*_positions.size(), _positions.data(), GL_STATIC_DRAW);

    glBindBuffer(GL_ARRAY_BUFFER, _vbo[1]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vector3f)*_normals.size(), _normals.data(), GL_STATIC_DRAW);

    glGenBuffers(1, &_indicesBuffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _indicesBuffer);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(Vector3i)*_indices.size(), _indices.data(),  GL_STATIC_DRAW);

    glBindVertexArray(0);

    _ready = true;
}

void Mesh::specifyVertexData(Shader *shader)
{
    glBindBuffer(GL_ARRAY_BUFFER, _vbo[0]);
    int pos_loc = shader->getAttribLocation("vtx_position");
    glEnableVertexAttribArray(pos_loc);
    glVertexAttribPointer(pos_loc, 3, GL_FLOAT, GL_FALSE, sizeof(Vector3f), (void*)0);

    glBindBuffer(GL_ARRAY_BUFFER, _vbo[1]);
    int normal_loc = shader->getAttribLocation("vtx_normal");
    if(normal_loc>=0){
        glEnableVertexAttribArray(normal_loc);
        glVertexAttribPointer(normal_loc, 3, GL_FLOAT, GL_FALSE, sizeof(Vector3f), (void*)0);
    }
}

void Mesh::display(Shader *shader)
{
    if (!_ready)
        init();

    glBindVertexArray(_vao);

    specifyVertexData(shader);

    glDrawElements(GL_TRIANGLES, _indices.size()*3,  GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
}

void Mesh::subdivide()
{
}
